#include "tree.h"
#include "util.h"
#include <stdio.h>
#include <sys/types.h> /* OM.h needs this - uck */
#include "OM.h"
#include <stdlib.h>
#include <string.h>

struct OMObjectNode *buildobject(OMdev dev);
struct OMIntegerNode *buildint32(OMdev dev);
struct OMIntegerNode *buildbigint(OMdev dev);
struct OMFloatingPNode *buildfloat64(OMdev dev);
struct OMByteArrayNode *buildbytearray(OMdev dev);
struct OMVariableNode *buildvar(OMdev dev);
struct OMStringNode *buildstring(OMdev dev);
struct OMWCStringNode *buildwcstring(OMdev dev);
struct OMSymbolNode *buildsymbol(OMdev dev);
struct OMApplicationNode *buildapp(OMdev dev);
struct OMAttributionNode *buildattr(OMdev dev);
struct OMErrorNode *builderror(OMdev dev);
struct OMBindingNode *buildbind(OMdev dev);
struct OMAtpNode *buildatp(OMdev dev);

struct OMObjectNode *buildtree(OMdev dev)
{
	OMtokenType type;
	struct OMObjectNode *tree;
	OMbool oldcomment;

	/* We're not interested in comments */
	oldcomment = OMignoreComment(dev,OMtrue);

	/* check for the start of the object */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenObject )
		die("Expecting Object got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetObject(dev));

	/* build the tree */
	tree = buildobject(dev);

	/* check for the closing object */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenEndObject )
		die("Expecting EndObject got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetEndObject(dev));

	/* set comments back the way they were and exit */
	OMignoreComment(dev,oldcomment);
	return tree;
}

struct OMObjectNode *buildobject(OMdev dev)
{
	OMtokenType type;
	struct OMObjectNode *obj;

	checkstatus(OMgetType(dev,&type));

	if( (obj = malloc(sizeof(*obj))) == NULL)
		die("Unable to malloc space for OMObjectNode.\n");
	
	switch(type) {
	case OMtokenInt32:
		obj->type = OMInteger;
		obj->node = buildint32(dev);
		break;
	case OMtokenBigInt:
		obj->type = OMInteger;
		obj->node = buildbigint(dev);
		break;
	case OMtokenFloat64:
		obj->type = OMFloatingP;
		obj->node = buildfloat64(dev);
		break;
	case OMtokenByteArray:
		obj->type = OMByteArray;
		obj->node = buildbytearray(dev);
		break;
	case OMtokenVar:
		obj->type = OMVariable;
		obj->node = buildvar(dev);
		break;
	case OMtokenString:
		obj->type = OMString;
		obj->node = buildstring(dev);
		break;
	case OMtokenWCString:
		obj->type = OMWCString;
		obj->node = buildwcstring(dev);
		break;
	case OMtokenSymbol:
		obj->type = OMSymbol;
		obj->node = buildsymbol(dev);
		break;
	case OMtokenApp:
		obj->type = OMApplication;
		obj->node = buildapp(dev);
		break;
	case OMtokenAttr:
		obj->type = OMAttribution;
		obj->node = buildattr(dev);
		break;
	case OMtokenError:
		obj->type = OMError;
		obj->node = builderror(dev);
		break;
	case OMtokenBind:
		obj->type = OMBinding;
		obj->node = buildbind(dev);
		break;
	default:
		die("Got %s type at start of an object!\n",
		    OMtokenTypeToString(type));
		break;
	}

	return obj;
}

struct OMIntegerNode *buildint32(OMdev dev)
{
	struct OMIntegerNode *obj;
	OMtokenType type;
	char buf[sizeof(int)*3+2]; /* sign, nul */
	int i;

	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenInt32 )
		die("Expecting Int32 got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetInt32(dev,&i));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMIntegerNode.\n");

	sprintf(buf,"%d",i);
	if( (obj->value = strdup(buf)) == NULL)
		die("Couldn't allocate memory for value of Integer.\n");

	return obj;
}


struct OMIntegerNode *buildbigint(OMdev dev)
{
	struct OMIntegerNode *obj;
	OMtokenType type;
	char *data;
	int sign;
	int len;
	OMbigIntType fmt;
	char *buf;

	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenBigInt )
		die("Expecting BigInt got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetBigInt(dev,&data,&len,&sign,&fmt));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMIntegerNode.\n");

	if( (buf = malloc(len + 3)) == NULL ) /* sign,x,null */
		die("Couldn't allocate memory for value of Integer.\n");

	strcpy(buf,"");
	if( sign == -1 )
		strcat(buf,"-");
	if( fmt == OMbigIntBase16 )
		strcat(buf,"x");
	strcat(buf,data);
	free(data);
	obj->value = buf;

	return obj;
}


struct OMFloatingPNode *buildfloat64(OMdev dev)
{
	struct OMFloatingPNode *obj;
	OMtokenType type;
	char buf[40]; /* 30,e,-,exponent,nul */
	double d;

	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenFloat64 )
		die("Expecting Float64 got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetFloat64(dev,&d));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMFloatingPNode.\n");

	sprintf(buf,"%.30g",d);
	if( (obj->value = strdup(buf)) == NULL)
		die("Couldn't allocate memory for value of Float.\n");

	return obj;
}


struct OMByteArrayNode *buildbytearray(OMdev dev)
{
	struct OMByteArrayNode *obj;
	OMtokenType type;
	char *data;
	int len;

	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenByteArray )
		die("Expecting ByteArray got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetByteArray(dev,&data,&len));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMByteArrayNode.\n");

	obj->array = data;
	obj->len = len;

	return obj;
}


struct OMVariableNode *buildvar(OMdev dev)
{
	struct OMVariableNode *obj;
	OMtokenType type;
	char *var;

	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenVar )
		die("Expecting Var got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetVar(dev,&var));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMVariableNode.\n");

	obj->varname = var;
	obj->unique_name = NULL;

	return obj;
}


struct OMStringNode *buildstring(OMdev dev)
{
	struct OMStringNode *obj;
	OMtokenType type;
	char *str;

	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenString )
		die("Expecting String got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetString(dev,&str));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMStringNode.\n");

	obj->text = str;

	return obj;
}


struct OMWCStringNode *buildwcstring(OMdev dev)
{
	struct OMWCStringNode *obj;
	OMtokenType type;
	OMUCS2 *str;

	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenWCString )
		die("Expecting WCString got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetWCString(dev,&str));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMWCStringNode.\n");

	obj->wctext = str;

	return obj;
}


struct OMSymbolNode *buildsymbol(OMdev dev)
{
	struct OMSymbolNode *obj;
	OMtokenType type;
	char *cd,*name;

	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenSymbol )
		die("Expecting Symbol got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetSymbol(dev,&cd,&name));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMSymbolNode.\n");

	obj->cd = cd;
	obj->name = name;

	return obj;
}


struct OMApplicationNode *buildapp(OMdev dev)
{
	struct OMApplicationNode *obj;
	OMtokenType type;
	int n;

	/* <OMA> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenApp )
		die("Expecting App got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetApp(dev));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMApplicationNode.\n");

	obj->head = buildobject(dev);
	if( (obj->args = malloc(sizeof(*(obj->args)))) == NULL )
		die("Couldn't allocate memory for Application args list.\n");

	checkstatus(OMgetType(dev,&type));
	for(n = 0; type != OMtokenEndApp; n++) {
		if( (obj->args = realloc(obj->args,sizeof(*(obj->args)) * (n+2))) == NULL )
			die("Couldn't expand Application args to %d.\n", n);
		(obj->args)[n] = buildobject(dev);
		checkstatus(OMgetType(dev,&type));
	}
	(obj->args)[n] = NULL;

	/* </OMA> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenEndApp )
		die("Expecting App got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetEndApp(dev));

	return obj;
}


struct OMAttributionNode *buildattr(OMdev dev)
{
	struct OMAttributionNode *obj;
	OMtokenType type;
	int n;

	/* <OMATTR> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenAttr )
		die("Expecting Attr got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetAttr(dev));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMAttributionNode.\n");

	/* <OMATP> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenAtp )
		die("Expecting Atp got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetAtp(dev));
	
	if( (obj->atps = malloc(sizeof(*(obj->atps)))) == NULL )
		die("Couldn't allocate memory for Attribution atp list.\n");

	checkstatus(OMgetType(dev,&type));
	for(n = 0; type != OMtokenEndAtp; n++) {
		if( (obj->atps = realloc(obj->atps,sizeof(*(obj->atps)) * (n+2))) == NULL )
			die("Couldn't expand Attribution atps to %d.\n", n);
		(obj->atps)[n] = buildatp(dev);
		checkstatus(OMgetType(dev,&type));
	}
	(obj->atps)[n] = NULL;

	/* </OMATP> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenEndAtp )
		die("Expecting EndAtp got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetEndAtp(dev));

	obj->object = buildobject(dev);

	/* </OMATTR> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenEndAttr )
		die("Expecting EndAttr got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetEndAttr(dev));

	return obj;
}


struct OMErrorNode *builderror(OMdev dev)
{
	struct OMErrorNode *obj;
	OMtokenType type;
	int n;

	/* <OME> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenError )
		die("Expecting Error got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetError(dev));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMErrorNode.\n");

	obj->symbol = buildsymbol(dev);

	if( (obj->args = malloc(sizeof(*(obj->args)))) == NULL )
		die("Couldn't allocate memory for Error args list.\n");

	checkstatus(OMgetType(dev,&type));
	for(n = 0; type != OMtokenEndError; n++) {
		if( (obj->args = realloc(obj->args,sizeof(*(obj->args)) * (n+2))) == NULL )
			die("Couldn't expand Error args to %d.\n", n);
		(obj->args)[n] = buildobject(dev);
		checkstatus(OMgetType(dev,&type));
	}
	(obj->args)[n] = NULL;

	/* </OME> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenEndError )
		die("Expecting EndError got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetEndError(dev));

	return obj;
}


struct OMBindingNode *buildbind(OMdev dev)
{
	struct OMBindingNode *obj;
	struct OMObjectNode *t;
	OMtokenType type;
	int n;

	/* <OMBIND> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenBind )
		die("Expecting Bind got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetBind(dev));

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMBindingNode.\n");

	obj->binder = buildobject(dev);

	/* <OMBVAR> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenBVar )
		die("Expecting BVar got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetBVar(dev));

	if( (obj->vars = malloc(sizeof(*(obj->vars)))) == NULL )
		die("Couldn't allocate memory for Bind vars list.\n");

	checkstatus(OMgetType(dev,&type));
	for(n = 0; type != OMtokenEndBVar; n++) {
		if( (obj->vars = realloc(obj->vars,sizeof(*(obj->vars)) * (n+2))) == NULL )
			die("Couldn't expand Bins vars to %d.\n", n);
		(obj->vars)[n] = buildobject(dev);
		/*
		 * Now check we actually got a variable, or an attributed var.
		 */
		t = (obj->vars)[n];
		while( t->type == OMAttribution )
			t = ((struct OMAttributionNode *)t->node)->object;
		if( t->type != OMVariable )
			die("Binding something that isn't a variable.\n", n);
		checkstatus(OMgetType(dev,&type));
	}
	(obj->vars)[n] = NULL;

	/* </OMBVAR> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenEndBVar )
		die("Expecting EndBVar got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetEndBVar(dev));

	obj->body = buildobject(dev);

	/* </OMBIND> */
	checkstatus(OMgetType(dev,&type));
	if( type != OMtokenEndBind )
		die("Expecting EndBind got a %s\n",OMtokenTypeToString(type));
	checkstatus(OMgetEndBind(dev));

	return obj;
}

struct OMAtpNode *buildatp(OMdev dev)
{
	struct OMAtpNode *obj;

	if( (obj = malloc(sizeof(*obj))) == NULL )
		die("Couldn't allocate memory for OMAtpNode.\n");

	obj->attribute = buildsymbol(dev);
	obj->value = buildobject(dev);

	return obj;
}
