#include "tree.h"
#include "util.h"
#include <stdlib.h>

void freeobject(struct OMObjectNode *obj);
void freeinteger(struct OMIntegerNode *obj);
void freesymbol(struct OMSymbolNode *obj);
void freevariable(struct OMVariableNode *obj);
void freefloatingp(struct OMFloatingPNode *obj);
void freestring(struct OMStringNode *obj);
void freewcstring(struct OMWCStringNode *obj);
void freebytearray(struct OMByteArrayNode *obj);
void freeapplication(struct OMApplicationNode *obj);
void freebinding(struct OMBindingNode *obj);
void freeattribution(struct OMAttributionNode *obj);
void freeerror(struct OMErrorNode *obj);
void freeatp(struct OMAtpNode *obj);

void freetree(struct OMObjectNode *tree)
{
	if( tree == NULL ) return;

	freeobject(tree);
}

void freeobject(struct OMObjectNode *object)
{
	if( object == NULL ) return;

	switch( object->type ) {
	case OMInteger:
		freeinteger(object->node);
		break;
	case OMSymbol:
		freesymbol(object->node);
		break;
	case OMVariable:
		freevariable(object->node);
		break;
	case OMFloatingP:
		freefloatingp(object->node);
		break;
	case OMString:
		freestring(object->node);
		break;
	case OMWCString:
		freewcstring(object->node);
		break;
	case OMByteArray:
		freebytearray(object->node);
		break;
	case OMApplication:
		freeapplication(object->node);
		break;
	case OMBinding:
		freebinding(object->node);
		break;
	case OMAttribution:
		freeattribution(object->node);
		break;
	case OMError:
		freeerror(object->node);
		break;
	default:
		die("Unknown object type %d.\n",object->type);
		break;
	}

	object->node = NULL;
	free(object);
}


void freeinteger(struct OMIntegerNode *obj)
{
	if( obj == NULL ) return;

	free(obj->value);
	obj->value = NULL;
	free(obj);
}


void freesymbol(struct OMSymbolNode *obj)
{
	if( obj == NULL ) return;

	free(obj->name);
	obj->name = NULL;

	free(obj->cd);
	obj->cd = NULL;

	free(obj);
}


void freevariable(struct OMVariableNode *obj)
{
	if( obj == NULL ) return;

	free(obj->varname);
	obj->varname = NULL;

	free(obj->unique_name);
	obj->unique_name = NULL;

	free(obj);
}


void freefloatingp(struct OMFloatingPNode *obj)
{
	if( obj == NULL ) return;

	free(obj->value);
	obj->value = NULL;

	free(obj);
}


void freestring(struct OMStringNode *obj)
{
	if( obj == NULL ) return;

	free(obj->text);
	obj->text = NULL;

	free(obj);
}


void freewcstring(struct OMWCStringNode *obj)
{
	if( obj == NULL ) return;

	free(obj->wctext);
	obj->wctext = NULL;

	free(obj);
}


void freebytearray(struct OMByteArrayNode *obj)
{
	if( obj == NULL ) return;

	free(obj->array);
	obj->array = NULL;

	free(obj);
}


void freeapplication(struct OMApplicationNode *obj)
{
	int n;

	if( obj == NULL ) return;

	freeobject(obj->head);
	obj->head = NULL;

	for( n = 0; (obj->args)[n] != NULL; n++ ) {
		freeobject((obj->args)[n]);
		(obj->args)[n] = NULL;
	}

	free(obj->args);
	obj->args = NULL;

	free(obj);
}


void freebinding(struct OMBindingNode *obj)
{
	int n;

	if( obj == NULL ) return;

	freeobject(obj->binder);
	obj->binder = NULL;

	for( n = 0; (obj->vars)[n] != NULL; n++ ) {
		freeobject((obj->vars)[n]);
		(obj->vars)[n] = NULL;
	}

	free(obj->vars);
	obj->vars = NULL;

	freeobject(obj->body);
	obj->body = NULL;

	free(obj);
}


void freeattribution(struct OMAttributionNode *obj)
{
	int n;

	if( obj == NULL ) return;

	for( n = 0; (obj->atps)[n] != NULL; n++ ) {
		freeatp((obj->atps)[n]);
		(obj->atps)[n] = NULL;
	}

	free(obj->atps);
	obj->atps = NULL;

	freeobject(obj->object);
	obj->object = NULL;

	free(obj);
}


void freeerror(struct OMErrorNode *obj)
{
	int n;

	if( obj == NULL ) return;

	freesymbol(obj->symbol);
	obj->symbol = NULL;

	for( n = 0; (obj->args)[n] != NULL; n++ ) {
		freeobject((obj->args)[n]);
		(obj->args)[n] = NULL;
	}

	free(obj->args);
	obj->args = NULL;

	free(obj);
}


void freeatp(struct OMAtpNode *obj)
{
	freesymbol(obj->attribute);
	obj->attribute = NULL;

	freeobject(obj->value);
	obj->value = NULL;
}
