#include "tree.h"
#include "util.h"
#include "perltree.h"
#include <stdio.h>
#include <ctype.h>

void perlobject(FILE *fp,struct OMObjectNode *obj);
void perlinteger(FILE *fp,struct OMIntegerNode *obj);
void perlsymbol(FILE *fp,struct OMSymbolNode *obj);
void perlvariable(FILE *fp,struct OMVariableNode *obj);
void perlfloatingp(FILE *fp,struct OMFloatingPNode *obj);
void perlstring(FILE *fp,struct OMStringNode *obj);
void perlwcstring(FILE *fp,struct OMWCStringNode *obj);
void perlbytearray(FILE *fp,struct OMByteArrayNode *obj);
void perlapplication(FILE *fp,struct OMApplicationNode *obj);
void perlbinding(FILE *fp,struct OMBindingNode *obj);
void perlattribution(FILE *fp,struct OMAttributionNode *obj);
void perlerror(FILE *fp,struct OMErrorNode *obj);
void perlatp(FILE *fp,struct OMAtpNode *obj);
void perlquotedstring(FILE *fp,char *string);

static int level;
static void uplevel(FILE *fp) { level++; fflush(fp);}
static void downlevel(FILE *fp) { level--; fflush(fp);}
static void indent(FILE *fp) { int n; for( n = 0; n < level; n++ ) fprintf(fp,"\t"); fflush(fp);}

void perltree(FILE *fp,struct OMObjectNode *tree)
{
	level = 1;

	perlobject(fp,tree);
}

void perlobject(FILE *fp,struct OMObjectNode *object)
{
	switch( object->type ) {
	case OMInteger:
		perlinteger(fp,object->node);
		break;
	case OMSymbol:
		perlsymbol(fp,object->node);
		break;
	case OMVariable:
		perlvariable(fp,object->node);
		break;
	case OMFloatingP:
		perlfloatingp(fp,object->node);
		break;
	case OMString:
		perlstring(fp,object->node);
		break;
	case OMWCString:
		perlwcstring(fp,object->node);
		break;
	case OMByteArray:
		perlbytearray(fp,object->node);
		break;
	case OMApplication:
		perlapplication(fp,object->node);
		break;
	case OMBinding:
		perlbinding(fp,object->node);
		break;
	case OMAttribution:
		perlattribution(fp,object->node);
		break;
	case OMError:
		perlerror(fp,object->node);
		break;
	default:
		die("Unknown object type %d.\n",object->type);
		break;
	}
}


void perlinteger(FILE *fp,struct OMIntegerNode *obj)
{
	fprintf(fp,"{\n");
	uplevel(fp);

	indent(fp); fprintf(fp,"'type'  => 'integer',\n");
	indent(fp); fprintf(fp,"'value' => '%s',\n",obj->value);

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlsymbol(FILE *fp,struct OMSymbolNode *obj)
{
	fprintf(fp,"{\n");
	uplevel(fp);

	indent(fp); fprintf(fp,"'type'  => 'symbol',\n");
	indent(fp); fprintf(fp,"'name'  => '%s',\n", obj->name);
	indent(fp); fprintf(fp,"'cd'    => '%s',\n", obj->cd);

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlvariable(FILE *fp,struct OMVariableNode *obj)
{
	fprintf(fp,"{\n");
	uplevel(fp);

	indent(fp); fprintf(fp,"'type'  => 'variable',\n");
	indent(fp); fprintf(fp,"'name'  => ");
	perlquotedstring(fp,obj->varname);
	fprintf(fp,",\n");
	if( obj->unique_name != NULL ) {
		indent(fp); fprintf(fp,"'u_name'=> ");
		perlquotedstring(fp,obj->unique_name);
		fprintf(fp,",\n");
	}

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlfloatingp(FILE *fp,struct OMFloatingPNode *obj)
{
	fprintf(fp,"{\n");
	uplevel(fp);

	indent(fp); fprintf(fp,"'type'  => 'float',\n");
	indent(fp); fprintf(fp,"'value' => '%s',\n",obj->value);

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlstring(FILE *fp,struct OMStringNode *obj)
{
	fprintf(fp,"{\n");
	uplevel(fp);

	indent(fp); fprintf(fp,"'type'  => 'string',\n");
	indent(fp); fprintf(fp,"'value' => ");
	perlquotedstring(fp,obj->text);
	fprintf(fp,",\n");

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlwcstring(FILE *fp,struct OMWCStringNode *obj)
{
	int n;
	unsigned int val;

	fprintf(fp,"{\n");
	uplevel(fp);

	indent(fp); fprintf(fp,"'type'  => 'string',\n");
	indent(fp); fprintf(fp,"'value'  => pack 'H*', '");
	/* utf8 encode the string */
	for( n = 0 ; (obj->wctext)[n] != 0 ; n++ ) {
		val = (obj->wctext)[n];
		if( val <= 0x007f )
			fprintf(fp,"%02x",
				0x00 | (val&0x007f)>>0);
		if( (obj->wctext)[n] <= 0x07ff )
			fprintf(fp,"%02x%02x",
				0xc0 | ((val&0x07c0)>>6),
				0x80 | ((val&0x003f)>>0));
		else
			fprintf(fp,"%02x%02x%02x",
				0xe0 | ((val&0xf000)>>12),
				0x80 | ((val&0x0fc0)>>6),
				0x80 | ((val&0x003f)>>0));
	}
	fprintf(fp,"',\n");

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlbytearray(FILE *fp,struct OMByteArrayNode *obj)
{
	int n;

	fprintf(fp,"{\n");
	uplevel(fp);

	indent(fp); fprintf(fp,"'type'  => 'bytearray',\n");

	indent(fp); fprintf(fp,"'value'  => pack 'H*', '");
	for( n = 0 ; n < obj->len ; n++ )
		fprintf(fp,"%02x",(obj->array)[n]);
	fprintf(fp,"',\n");

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlapplication(FILE *fp,struct OMApplicationNode *obj)
{
	int n;

	fprintf(fp,"{\n");
	uplevel(fp);
	indent(fp); fprintf(fp,"'type'  => 'application',\n");

	indent(fp); fprintf(fp,"'head'  => ");
	perlobject(fp,obj->head);
	fprintf(fp,",\n");

	indent(fp); fprintf(fp,"'args'  => [");
	for( n = 0; (obj->args)[n] != NULL; n++ ) {
		if ( n > 0 ) indent(fp);
		perlobject(fp,(obj->args)[n]);
		fprintf(fp,",\n");
	}
	indent(fp); fprintf(fp,"],\n");

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlbinding(FILE *fp,struct OMBindingNode *obj)
{
	int n;

	fprintf(fp,"{\n");
	uplevel(fp);
	indent(fp); fprintf(fp,"'type'  => 'binding',\n");

	indent(fp); fprintf(fp,"'binder'=> ");
	perlobject(fp,obj->binder);
	fprintf(fp,",\n");

	indent(fp); fprintf(fp,"'vars'  => [");
	for( n = 0; (obj->vars)[n] != NULL; n++ ) {
		if ( n > 0 ) indent(fp);
		perlobject(fp,(obj->vars)[n]);
		fprintf(fp,",\n");
	}
	indent(fp); fprintf(fp,"],\n");

	indent(fp); fprintf(fp,"'body'  => ");
	perlobject(fp,obj->body);
	fprintf(fp,",\n");

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlattribution(FILE *fp,struct OMAttributionNode *obj)
{
	int n;

	fprintf(fp,"{\n");
	uplevel(fp);
	indent(fp); fprintf(fp,"'type'  => 'attribution',\n");

	indent(fp); fprintf(fp,"'object'=> ");
	perlobject(fp,obj->object);
	fprintf(fp,",\n");

	indent(fp); fprintf(fp,"'attrs' => [");
	for( n = 0; (obj->atps)[n] != NULL; n++ ) {
		if ( n > 0 ) indent(fp);
		perlatp(fp,(obj->atps)[n]);
		fprintf(fp,",\n");
	}
	indent(fp); fprintf(fp,"],\n");

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlerror(FILE *fp,struct OMErrorNode *obj)
{
	int n;

	fprintf(fp,"{\n");
	uplevel(fp);
	indent(fp); fprintf(fp,"'type'  => 'error',\n");

	indent(fp); fprintf(fp,"'symbol'=> \n");
	perlsymbol(fp,obj->symbol);
	fprintf(fp,",\n");

	indent(fp); fprintf(fp,"'args'  => [\n");
	for( n = 0; (obj->args)[n] != NULL; n++ ) {
		if ( n > 0 ) indent(fp);
		perlobject(fp,(obj->args)[n]);
		fprintf(fp,",\n");
	}
	indent(fp); fprintf(fp,"],\n");

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}


void perlatp(FILE *fp,struct OMAtpNode *obj)
{
	fprintf(fp,"{\n");
	uplevel(fp);
	indent(fp); fprintf(fp,"'type'       => 'atp',\n");

	indent(fp); fprintf(fp,"'attribute'  => ");
	perlsymbol(fp,obj->attribute);
	fprintf(fp,",\n");

	indent(fp); fprintf(fp,"'value'      => ");
	perlobject(fp,obj->value);
	fprintf(fp,",\n");

	downlevel(fp);
	indent(fp); fprintf(fp,"}");
}

void perlquotedstring(FILE *fp,char *string)
{
	int n;
	int single_quote = 0;
	int double_quote = 0;
	int slash = 0;
	int unprintable = 0;

	for( n = 0 ; string[n] != '\0' ; n++ ) {
		if( string[n] == '\'' ) single_quote = 1;
		if( string[n] == '"' ) double_quote = 1;
		if( string[n] == '/' ) slash = 1;
		if( !isprint(string[n]) ) unprintable = 1;
	}

	if( unprintable || (single_quote && double_quote && slash) ) {
		fprintf(fp,"pack 'H*', '");
		for( n = 0 ; string[n] != '\0' ; n++ )
			fprintf(fp,"%02x",string[n]);
		fprintf(fp,"'");
		return;
	}

	if( !single_quote ) {
		fprintf(fp,"'%s'", string);
	} else if( !slash ) {
		fprintf(fp,"q/%s/", string);
	} else if( !double_quote ) {
		fprintf(fp,"q\"%s\"", string);
	} else {
		die("Couldn't quote string in perlquotedstring!");
	}
}
