#include "tree.h"
#include "util.h"
#include "freetree.h"
#include <string.h>
#include <stdlib.h>

void tidyobject(struct OMObjectNode *obj);
void tidyinteger(struct OMIntegerNode *obj);
void tidysymbol(struct OMSymbolNode *obj);
void tidyvariable(struct OMVariableNode *obj);
void tidyfloatingp(struct OMFloatingPNode *obj);
void tidystring(struct OMStringNode *obj);
void tidywcstring(struct OMWCStringNode *obj);
void tidybytearray(struct OMByteArrayNode *obj);
void tidyapplication(struct OMApplicationNode *obj);
void tidybinding(struct OMBindingNode *obj);
void tidyattribution(struct OMAttributionNode *obj);
void tidyerror(struct OMErrorNode *obj);
void tidyatp(struct OMAtpNode *obj);

char *getnewvarname();
void weakbind(struct OMVariableNode *obj);
void pushbound(struct OMVariableNode *obj);
void popbound(struct OMVariableNode *obj);
char *bound_name(char *varname);
void cleanstack(void);

void tidytree(struct OMObjectNode *tree)
{
	tidyobject(tree);
	cleanstack();
}

void tidyobject(struct OMObjectNode *object)
{

	switch( object->type ) {
	case OMInteger:
		tidyinteger(object->node);
		break;
	case OMSymbol:
		tidysymbol(object->node);
		break;
	case OMVariable:
		tidyvariable(object->node);
		break;
	case OMFloatingP:
		tidyfloatingp(object->node);
		break;
	case OMString:
		tidystring(object->node);
		break;
	case OMWCString:
		tidywcstring(object->node);
		break;
	case OMByteArray:
		tidybytearray(object->node);
		break;
	case OMApplication:
		tidyapplication(object->node);
		break;
	case OMBinding:
		tidybinding(object->node);
		break;
	case OMAttribution:
		tidyattribution(object->node);
		break;
	case OMError:
		tidyerror(object->node);
		break;
	default:
		die("Unknown object type %d.\n",object->type);
		break;
	}
}


void tidyinteger(struct OMIntegerNode *obj)
{
	if( obj == NULL ) die("Void node"); /* stop gcc warning */
}


void tidysymbol(struct OMSymbolNode *obj)
{
	if( obj == NULL ) die("Void node"); /* stop gcc warning */
}


void tidyvariable(struct OMVariableNode *obj)
{
	char *newvarname;

	if( (newvarname = bound_name(obj->varname)) != NULL) {
		if( (obj->unique_name = strdup(newvarname)) == NULL )
			die("Couldn't strdup newvarname!\n");
	} else {
		weakbind(obj);
		obj->unique_name = getnewvarname();
	}

}


void tidyfloatingp(struct OMFloatingPNode *obj)
{
	if( obj == NULL ) die("Void node"); /* stop gcc warning */
}


void tidystring(struct OMStringNode *obj)
{
	if( obj == NULL ) die("Void node"); /* stop gcc warning */
}


void tidywcstring(struct OMWCStringNode *obj)
{
	if( obj == NULL ) die("Void node"); /* stop gcc warning */
}


void tidybytearray(struct OMByteArrayNode *obj)
{
	if( obj == NULL ) die("Void node"); /* stop gcc warning */
}


void tidyapplication(struct OMApplicationNode *obj)
{
	int n;

	tidyobject(obj->head);

	for( n = 0 ; (obj->args)[n] != NULL; n++ )
		tidyobject((obj->args)[n]);
}


void tidybinding(struct OMBindingNode *obj)
{
	int n;
	struct OMObjectNode *t;

	tidyobject(obj->binder);

	for( n = 0; (obj->vars)[n] != NULL; n++ ) {
		t = (obj->vars)[n];
		while( t->type == OMAttribution )
			t = ((struct OMAttributionNode *)t->node)->object;
		if( t->type != OMVariable )
			die("Tidying a bound something that isn't a variable.\n", n);
		pushbound(t->node);
		tidyobject((obj->vars)[n]);
	}

	tidyobject(obj->body);

	for( n--; n >= 0; n-- ) {
		t = (obj->vars)[n];
		while( t->type == OMAttribution )
			t = ((struct OMAttributionNode *)t->node)->object;
		if( t->type != OMVariable )
			die("Tidying a bound something that isn't a variable.\n", n);
		popbound(t->node);
	}
}


void tidyattribution(struct OMAttributionNode *obj)
{
	int n,n2;
	struct OMAttributionNode *obj2;
	struct OMObjectNode *old;

	for( n = 0; (obj->atps)[n] != NULL; n++ )
		tidyatp((obj->atps)[n]);

	tidyobject(obj->object);

	/* Fold attribution of attribution into attribution */
	if( obj->object->type == OMAttribution) {
		obj2 = obj->object->node;
		obj->object->node = NULL;

		for( n2 = 0; (obj2->atps)[n2] != NULL; n2++ )
			;
		if( (obj->atps = realloc(obj->atps,sizeof(*(obj->atps))*(n+n2+1))) == NULL )
			die("Couldn't expand Attribution atps to %d.\n", n);
		for( n2 = 0; (obj2->atps)[n2] != NULL; n2++,n++ ) {
			(obj->atps)[n] = (obj2->atps)[n2];
			(obj2->atps)[n2] = NULL;
		}
		(obj->atps)[n] = NULL;

		old = obj->object;
		obj->object = obj2->object;
		obj2->object = NULL;
		freeobject(old); /* frees obj2 as old->node == obj2 */
	}
}


void tidyerror(struct OMErrorNode *obj)
{
	int n;

	tidysymbol(obj->symbol);

	for( n = 0; (obj->args)[n] != NULL; n++ )
		tidyobject((obj->args)[n]);
}


void tidyatp(struct OMAtpNode *obj)
{
	tidysymbol(obj->attribute);
	tidyobject(obj->value);
}

/*
 *	Functions to do with identifing which variables are the same.
 */

static long variable_count = 0;
static struct OMVariableNode **bound_stack = NULL;
static long bound_number = 0;
static struct OMVariableNode **weak_list = NULL;
static long weak_number = 0;

char *getnewvarname()
{
	char *newvarname;

	if( (newvarname = malloc(sizeof(bound_number)*3+6)) == NULL )
		die("Couldn't malloc space for newvarname\n");
	sprintf(newvarname,"omvar%ld",variable_count++);

	return newvarname;
}

void weakbind(struct OMVariableNode *obj)
{
	if( (weak_list = realloc(weak_list,sizeof(obj)*(weak_number+1))) == NULL )
		die("Couldn't allocate memory for weak_list!");

	weak_list[weak_number] = obj;
	weak_number++;
}

void pushbound(struct OMVariableNode *obj)
{
	if( (bound_stack = realloc(bound_stack,sizeof(obj)*(bound_number+1))) == NULL )
		die("Couldn't allocate memory for bound_stack!");

	bound_stack[bound_number] = obj;
	bound_number++;
}

void popbound(struct OMVariableNode *obj)
{
	if( bound_number <= 0 )
		die("Popping empty bound varibale stack!");

	bound_number--;

	if( bound_stack[bound_number] != obj )
		die("Popping incorrect variable!\n");

	bound_stack[bound_number] = NULL;
}

char *bound_name(char *varname)
{
	int n;

	for( n = bound_number-1; n >= 0 ; n-- )
		if( strcmp(varname,bound_stack[n]->varname) == 0 )
			return bound_stack[n]->unique_name;

	for( n = 0; n < weak_number ; n++ )
		if( strcmp(varname,weak_list[n]->varname) == 0 )
			return weak_list[n]->unique_name;

	return NULL;
}

void cleanstack(void)
{
	if( bound_number != 0 )
		die("Junk left on the bound variable stack!\n");

	free(bound_stack);
	bound_stack = NULL;

	weak_number = 0;
	free(weak_list);
	weak_list = NULL;
}
