#

#
# These are the drivers for formatting each type. They basically figure
# out if there is a custom formatter to call and if not they call the
# generic one. There is a further table for these for the compound types.
# The main functions you will want to call here are:
#
#	format_tree($rootnode)	Formats a whole tree as a string.
#	format_type($node)	Formats a note and returns the format info.
#	format_die("message")	Die, printing a trace of the tree so far.
#
# You'll also want to provide &generic_load_cdformatter to load the perl for
# the CD formatters, &generic_load_variableformatter to load perl for variable
# formatting and %generic_format_type an array of functions to do the formatting
#

sub format_die {
	my ($message) = @_;

	print STDERR $message,"\n";
	print STDERR "Node stack:\n";
	foreach $node (@nodestack) { &print_node($node) };
	die "Can't continue";
}

sub format_tree {
	my ($node) = @_;

	return &{$generic_format_type{'tree'}}($node);
}

%format_type = (
	'integer'     =>	\&format_integer,
	'symbol'      =>	\&format_symbol,
	'variable'    =>	\&format_variable,
	'float'       =>	\&format_float,
	'string'      =>	\&format_string,
	'bytearray'   =>	\&format_bytearray,
	'application' =>	\&format_application,
	'binding'     =>	\&format_binding,
	'attribution' =>	\&format_attribution,
	'error'       =>	\&format_error,
	'atp'         =>	\&format_atp,
);

sub format_type {
	my ($node) = @_;

	push(@nodestack,$node);
	if( !defined($format_type{$node->{'type'}}) ) {
		&format_die("Asked to format bad type.");
	}
	if( !defined($node->{'format'}) ) {
		$node->{'format'} = &{$format_type{$node->{'type'}}}($node);
	}
	pop(@nodestack);
	return $node->{'format'};
}

sub generic_format_type {
	my ($node) = @_;

	return &{$generic_format_type{$node->{'type'}}}($node);
}

sub format_integer {
	my ($node) = @_;

	return &generic_format_type($node);
}

sub format_symbol {
	my ($node) = @_;
	my $cd = $node->{'cd'};
	my $name = $node->{'name'};

	&generic_load_cdformatter($cd);

	if( defined($cdformatter{$cd}) ) {
		if(defined($cdformatter{$cd}->{$name}->{'symbol'})) {
			return &{$cdformatter{$cd}->{$name}->{'symbol'}}($node);
		} elsif(defined($cdformatter{$cd}->{'generic'})) {
			return &{$cdformatter{$cd}->{'generic'}}($node);
		}
	}

	return &generic_format_type($node);
}

sub format_variable {
	my ($node) = @_;
	my $name = $node->{'name'};
	my $result;

	&generic_load_variableformatter();
	if( &is_bound($name) && defined(&bound_format($name))) {
		return &bound_format($name);
	} elsif( defined($varaibleformatter{$name}->{'variable'}) ) {
		$result = &{$varaibleformatter{$name}->{'variable'}}($node);
	} else {
		$result = &generic_format_type($node);
	}

	bindformat($name,$result);
	return $result;
}

sub format_float {
	my ($node) = @_;

	return &generic_format_type($node);
}

sub format_string {
	my ($node) = @_;

	return &generic_format_type($node);
}

sub format_bytearray {
	my ($node) = @_;

	return &generic_format_type($node);
}

sub format_application {
	my ($node) = @_;
	my $head = $node->{'head'};

	if( defined($format_application_type{$head->{'type'}}) ) {
		return &{$format_application_type{$head->{'type'}}}($node);
	}

	return &generic_format_type($node);
}

sub format_binding {
	my ($node) = @_;
	my $binder = $node->{'binder'};

	if( defined($format_binding_type{$binder->{'type'}}) ) {
		return &{$format_binding_type{$binder->{'type'}}}($node);
	}

	return &generic_format_type($node);
}

#
# This is slightly complicated 'cos we have to make lists look like
# recursion. Maybe I should just turn it into a tree...
#

sub format_attribution {
	my ($node) = @_;
	my $object = $node->{'object'};
	my $attrs = $node->{'attrs'};
	my $atp;
	my $attribute;
	my $dummy;
	my $format;

	$format = undef;
	foreach $atp ( reverse @$attrs ) {
		$dummy = {
			'type' => 'attribution',
			'object' => $object,
			'attrs' => [ $atp ],
		};
		$attribute = $dummy->{'attrs'}->[0]->{'attribute'};
		if( defined($format_attribution_type{$attribute->{'type'}}) ) {
			$format = &{$format_attribution_type{$attribute->{'type'}}}($dummy);
		} else {
			$format = &generic_format_type($dummy);
		}
		$object->{'format'} = $format;
	}

	return $format;
}

sub format_error {
	my ($node) = @_;
	my $symbol = $node->{'symbol'};
	my $cd = $symbol->{'cd'};
	my $name = $symbol->{'name'};

	&generic_load_cdformatter($cd);

	if( defined($cdformatter{$cd}) ) {
		if(defined($cdformatter{$cd}->{$name}->{'error'})) {
			return &{$cdformatter{$cd}->{$name}->{'error'}}($node);
		} elsif(defined($cdformatter{$cd}->{'generic'})) {
			return &{$cdformatter{$cd}->{'generic'}}($node);
		}
	}

	return &generic_format_type($node);
}


################################################################################

#
# These are the drivers that check and see if the primary bit of a
# compound object knows how to be formatted.
#

%format_application_type = (
	'symbol'      =>	\&format_application_symbol,
	'variable'    =>	\&format_application_variable,
);

%format_binding_type = (
	'symbol'      =>	\&format_binding_symbol,
	'variable'    =>	\&format_binding_variable,
);


%format_attribution_type = (
	'symbol'      =>	\&format_attribution_symbol,
	'variable'    =>	\&format_attribution_variable,
);


sub format_application_symbol {
	my ($node) = @_;
	my $head = $node->{'head'};
	my $cd = $head->{'cd'};
	my $name = $head->{'name'};

	&generic_load_cdformatter($cd);

	if( defined($cdformatter{$cd}) ) {
		if(defined($cdformatter{$cd}->{$name}->{'application'})) {
			return &{$cdformatter{$cd}->{$name}->{'application'}}($node);
		} elsif(defined($cdformatter{$cd}->{'generic'})) {
			return &{$cdformatter{$cd}->{'generic'}}($node);
		}
	}

	return &generic_format_type($node);
}

sub format_application_variable {
	my ($node) = @_;
	my $head = $node->{'head'};
	my $name = $head->{'name'};

	&generic_load_variableformatter();
	if( defined($varaibleformatter{$name}->{'application'}) ) {
		return &{$varaibleformatter{$name}->{'application'}}($node);
	}

	return &generic_format_type($node);
}

sub format_binding_symbol {
	my ($node) = @_;
	my $binder = $node->{'binder'};
	my $cd = $binder->{'cd'};
	my $name = $binder->{'name'};

	&generic_load_cdformatter($cd);

	if( defined($cdformatter{$cd}) ) {
		if(defined($cdformatter{$cd}->{$name}->{'binding'})) {
			return &{$cdformatter{$cd}->{$name}->{'binding'}}($node);
		} elsif(defined($cdformatter{$cd}->{'generic'})) {
			return &{$cdformatter{$cd}->{'generic'}}($node);
		}
	}

	return &generic_format_type($node);
}

sub format_binding_variable {
	my ($node) = @_;
	my $binder = $node->{'binder'};
	my $name = $head->{'name'};

	&generic_load_variableformatter();
	if( defined($varaibleformatter{$name}->{'binding'}) ) {
		return &{$varaibleformatter{$name}->{'binding'}}($node);
	}

	return &generic_format_type($node);
}

sub format_attribution_symbol {
	my ($node) = @_;
	my $object = $node->{'attrs'}->[0]->{'attribute'};
	my $cd = $object->{'cd'};
	my $name = $object->{'name'};

	&generic_load_cdformatter($cd);

	if( defined($cdformatter{$cd}) ) {
		if(defined($cdformatter{$cd}->{$name}->{'attribution'})) {
			return &{$cdformatter{$cd}->{$name}->{'attribution'}}($node);
		} elsif(defined($cdformatter{$cd}->{'generic'})) {
			return &{$cdformatter{$cd}->{'generic'}}($node);
		}
	}

	return &generic_format_type($node);
}

sub format_attribution_variable {
	my ($node) = @_;
	my $object = $node->{'object'};
	my $name = $head->{'name'};

	&generic_load_variableformatter();
	if( defined($varaibleformatter{$name}->{'attribution'}) ) {
		return &{$varaibleformatter{$name}->{'attribution'}}($node);
	}

	return &generic_format_type($node);
}

1;
