# Generic formatters for LaTeX

require "tex/-useful.pl";

# We return a structure containing a normal formatting, a simple formatting
# and a left and right bracket for the object. eg:
#
# <OMBIND> <OMS name="lambda" cd="fns"/> 
#	<OMBVAR> <OMV name="x"/> </OMBVAR>
#	<OMA> <OMS name="plus" cd="arith1"/> <OMV name="x"/> <OMI>2</OMI> </OMA>
# </OMBIND>
#
# might end up as:
#
# {
#	'fmt' => '{\mbox{lambda}.{x} \rightarrow {x + 2}}',
#	'sfmt'=> '{x + 2}',
#	'lpren'  => '\left(',
#	'rpren'  => '\right)',
# }
#
#

%generic_format_type = (
	'tree'        =>	\&tex_format_tree,
	'integer'     =>	\&tex_format_integer,
	'symbol'      =>	\&tex_format_symbol,
	'variable'    =>	\&tex_format_variable,
	'float'       =>	\&tex_format_float,
	'string'      =>	\&tex_format_string,
	'bytearray'   =>	\&tex_format_bytearray,
	'application' =>	\&tex_format_application,
	'binding'     =>	\&tex_format_binding,
	'attribution' =>	\&tex_format_attribution,
	'error'       =>	\&tex_format_error,
	'atp'         =>	\&tex_format_atp,
);

sub tex_format_tree {
	my ($node) = @_;

	return &format_type($node)->{'fmt'};
}

sub tex_format_integer {
	my ($node) = @_;
	my $fmt;

	$fmt = $node->{'value'};

	return ({'fmt' => $fmt, 'sfmt' => $fmt,
		'lpren' => '', 'rpren' => ''});
}


sub tex_format_symbol {
	my ($node) = @_;
	my $fmt;

	$fmt = '\mbox{' . &tex_escape_text_string($node->{'name'}) . '}';

	return ({'fmt' => $fmt, 'sfmt' => $fmt,
		'lpren' => '\left(', 'rpren' => '\right)'});
}


sub tex_format_variable {
	my ($node) = @_;
	my $count = 0;
	my $fmt;
	my $name;

	$name = &tex_escape_math_string($node->{'name'});
	$fmt = $name;
	while( &format_used($fmt) ) { # If the format is used try subscripting
		$count++;
		$fmt = $name . '_{' . $count . '}';
	}

	return ({'fmt' => $fmt, 'sfmt' => $fmt,
		'lpren' => '', 'rpren' => ''});
}


sub tex_format_float {
	my ($node) = @_;
	my $value = $node->{'value'};
	my ($man, $exp);
	my $fmt;

	$value =~ s/^\s*\+\s*//;	# Remove leading space and +
	if( ($value =~ /[eE]/i) ) {	# If we have an exponent write as 10^e
		($man,$exp) = split(/[eE]/,$value);
		$exp =~ s/^\s*\+\s*//;
		$value = "$man \\times 10^{$exp}";
	}

	$fmt = $value;

	return ({'fmt' => $fmt, 'sfmt' => $fmt,
		'lpren' => '', 'rpren' => ''});
}


sub tex_format_string {
	my ($node) = @_;
	my $fmt;

	$fmt = '\mbox{' . &tex_escape_text_string($node->{'value'}) . '}';

	return ({'fmt' => $fmt, 'sfmt' => $fmt,
		'lpren' => '', 'rpren' => ''});
}


sub tex_format_bytearray {
	my ($node) = @_;
	my $fmt;

	$fmt = '\mbox{\texttt{0x' . unpack('H*',$node->{'value'}) . '}}';

	return ({'fmt' => $fmt, 'sfmt' => $fmt,
		'lpren' => '', 'rpren' => ''});
}


sub tex_format_application {
	my ($node) = @_;
	my $head = $node->{'head'};
	my $args = $node->{'args'};
	my $fhead;
	my @fargs;
	my $arg;
	my $fmt;
	
	$fhead = &tex_bracket_fmt(&format_type($head));

	foreach $arg ( @$args ) {
		push(@fargs, &format_type($arg)->{'sfmt'});
	}

	$fmt = $fhead . '\left(' . join(',',@fargs) . '\right)';

	return ({'fmt' => $fmt, 'sfmt' => $fmt,
		'lpren' => '\left(', 'rpren' => '\right)'});
}


sub tex_format_binding {
	my ($node) = @_;
	my $binder = $node->{'binder'};
	my $vars = $node->{'vars'};
	my $body = $node->{'body'};
	my ($fbinder, @fvars, $fbody);
	my $var;
	my $fmt;

	$fbinder = &format_type($binder)->{'fmt'};
	foreach $var ( @$vars ) {
		pushbind($var);
		push(@fvars, &format_type($var)->{'sfmt'});
	}
	$fbody = &format_type($body)->{'fmt'};
	foreach $var ( @$vars ) {
		popbind($var);
	}

	$fmt = $fbinder .  '.' . join(',',@fvars) . 
		' \rightarrow ' . $fbody ;

	return ({'fmt' => $fmt, 'sfmt' => $fbody,
		'lpren' => '\left(', 'rpren' => '\right)'});
}


sub tex_format_attribution {
	my ($node) = @_;
	my $object = $node->{'object'};
	my $attrs = $node->{'attrs'};
	my $atp;
	my $fmt;
	my $sfmt;

	$atp = $attrs->[0];
	$sfmt = &format_type($object)->{'sfmt'};
	$fmt = $sfmt .
		'\mbox{ has attribute }' .
		&format_type($atp->{'attribute'})->{'fmt'} .
		'\mbox{ with value }' .
		&format_type($atp->{'value'})->{'fmt'} .
		' , ' .
		&format_type($object)->{'fmt'};


	return ({'fmt' => $fmt, 'sfmt' => $sfmt,
		'lpren' => '\left(', 'rpren' => '\right)'});
}


sub tex_format_error {
	my ($node) = @_;
	my $symbol = $node->{'symbol'};
	my $args = $node->{'args'};
	my $fsymbol;
	my @fargs;
	my $arg;
	my $fmt;
	
	$fsymbol = &format_type($symbol)->{'fmt'};
	foreach $arg ( @$args ) {
		push(@fargs, &format_type($arg)->{'fmt'});
	}

	$fmt = '\mbox{ An error of type }' .  $fsymbol .
		'\mbox{ occured } \left(' . join(',',@fargs) .
		'\right)';

	return ({'fmt' => $fmt, 'sfmt' => $fsymbol,
		'lpren' => '', 'rpren' => ''});
}

sub generic_load_cdformatter {
	my ($cd) = @_;
	my $filename = "tex/$cd.pl";

	return if($loaded_cdformatter{$cd});

	die "Bad cd name '$cd'" unless( $cd =~ /^[a-z0-9A-Z_]+$/ );

	foreach $path ( @INC ) {
		if( -f "$path/$filename" ) {
			require $filename;
			$loaded_cdformatter{$cd} = 1;
			return;
		}
	}

	print STDERR "Unable to load $filename for cd $cd\n";
	$loaded_cdformatter{$cd} = 1;
}

sub generic_load_variableformatter {
	my $filename = "tex/-variables.pl";

	return if($loaded_variableformatter);

	foreach $path ( @INC ) {
		if( -f "$path/$filename" ) {
			require $filename;
			$loaded_variableformatter = 1;
			return;
		}
	}

	$loaded_variableformatter = 1;
}

%texmathescapes = (
	'\\' => '\\backslash',
	'~'  => '\\tilde{\\ }',
	'^'  => '\\hat{\\ }',
	'#'  => '\\#',
	'$'  => '\$',
	'%'  => '\%',
	'&'  => '\&',
	'_'  => '\_',
	'{'  => '\{',
	'}'  => '\}',
);

sub tex_escape_math_string {
	my ($string) = @_;

	$string =~ s/([\\\~\^\#\$\%\&\_\{\}])/\{$texmathescapes{$1}\}/g;

	return $string;
}

%textextescapes = (
	'\\' => '$\\backslash$',
	'~'  => '$\\tilde{\\ }$',
	'^'  => '$\\hat{\\ }$',
	'#'  => '\\#',
	'$'  => '\$',
	'%'  => '\%',
	'&'  => '\&',
	'_'  => '\_',
	'{'  => '\{',
	'}'  => '\}',
);

sub tex_escape_text_string {
	my ($string) = @_;

	$string =~ s/([\\\~\^\#\$\%\&\_\{\}])/\{$textextescapes{$1}\}/g;

	return $string;
}

1;
